#!/usr/bin/env python3
# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Script to import a RIO10 model into a kapture.
The RGB-D Dataset RIO10 data structure is defined here:
    https://github.com/WaldJohannaU/RIO10#dataformat

Sequence seq<scene_id>_01 is always the training sequence (with *.color.jpg, *.pose.txt and *.rendered.depth.png)
seqXX_02 is the validation sequence (with *.color.jpg, *.pose.txt and *.rendered.depth.png).
Please note that we do not provide the ground truth (including semantics) for our hidden test set
all seqXX_02+ are hidden sequences (only *.color.jpg and *.rendered.depth.png)
"""

import argparse
import logging
import os
import os.path as path
import re
import numpy as np
import quaternion
from PIL import Image
from tqdm import tqdm
# kapture
import path_to_kapture  # noqa: F401
from kapture.core.Sensors import SENSOR_TYPE_DEPTH_CAM
import kapture
import kapture.utils.logging
from kapture.io.structure import delete_existing_kapture_files
from kapture.io.csv import kapture_to_dir
import kapture.io.features
from kapture.io.records import TransferAction, import_record_data_from_dir_auto
from typing import List

logger = logging.getLogger('RIO10')

TIMESTAMP_SEQUENCE_BREAK = 999
DEPTH_TO_METER = 1.0e-3  # mm to meter

MAPPING_SEQUENCE_ID = 1
VALIDATION_SEQUENCE_ID = 2

POSE_SUFFIX = 'pose'
RGB_SUFFIX = 'color'
DEPTH_SUFFIX = 'rendered.depth'

WIDTH = 540
HEIGHT = 960
CAMERA_TYPE = kapture.CameraType.PINHOLE

INTRINSICS = {
    'seq01_01': [756.026, 756.832, 270.419, 492.889],
    'seq01_02': [756.026, 756.832, 270.418, 492.889],
    'seq01_03': [756.026, 756.832, 270.419, 492.889],
    'seq01_04': [758.792, 760.052, 266.208, 485.694],
    'seq01_05': [759.611, 760.513, 268.301, 491.445],
    'seq01_06': [759.611, 760.513, 268.301, 491.445],
    'seq02_01': [756.026, 756.832, 270.419, 492.889],
    'seq02_02': [758.792, 760.051, 266.208, 485.693],
    'seq02_03': [756.026, 756.832, 270.419, 492.889],
    'seq02_04': [756.026, 756.832, 270.419, 492.889],
    'seq02_05': [755.581, 756.525, 269.186, 488.945],
    'seq02_06': [759.507, 760.646, 268.339, 487.215],
    'seq02_07': [759.507, 760.646, 268.339, 487.215],
    'seq02_08': [756.026, 756.832, 270.419, 492.889],
    'seq03_01': [756.026, 756.832, 270.419, 492.889],
    'seq03_02': [756.026, 756.832, 270.418, 492.889],
    'seq03_03': [756.026, 756.832, 270.419, 492.889],
    'seq03_04': [756.026, 756.832, 270.419, 492.889],
    'seq03_05': [756.026, 756.832, 270.419, 492.889],
    'seq03_06': [758.792, 760.052, 266.208, 485.694],
    'seq03_07': [756.026, 756.832, 270.419, 492.889],
    'seq04_01': [758.792, 760.052, 266.208, 485.694],
    'seq04_02': [756.026, 756.832, 270.418, 492.889],
    'seq04_03': [756.026, 756.832, 270.419, 492.889],
    'seq04_04': [756.026, 756.832, 270.419, 492.889],
    'seq04_05': [756.026, 756.832, 270.419, 492.889],
    'seq04_06': [756.026, 756.832, 270.419, 492.889],
    'seq04_07': [756.026, 756.832, 270.419, 492.889],
    'seq04_08': [756.026, 756.832, 270.419, 492.889],
    'seq04_09': [756.026, 756.832, 270.419, 492.889],
    'seq04_10': [760.420, 761.683, 268.748, 486.778],
    'seq05_01': [756.026, 756.832, 270.419, 492.889],
    'seq05_02': [760.420, 761.683, 268.747, 486.778],
    'seq05_03': [756.026, 756.832, 270.419, 492.889],
    'seq05_04': [756.026, 756.832, 270.419, 492.889],
    'seq05_05': [758.792, 760.052, 266.208, 485.694],
    'seq06_01': [756.026, 756.832, 270.419, 492.889],
    'seq06_02': [756.026, 756.832, 270.418, 492.889],
    'seq06_03': [756.026, 756.832, 270.419, 492.889],
    'seq06_04': [756.026, 756.832, 270.419, 492.889],
    'seq06_05': [756.026, 756.832, 270.419, 492.889],
    'seq06_06': [756.026, 756.832, 270.419, 492.889],
    'seq06_07': [756.026, 756.832, 270.419, 492.889],
    'seq06_08': [756.026, 756.832, 270.419, 492.889],
    'seq06_09': [756.026, 756.832, 270.419, 492.889],
    'seq06_10': [756.026, 756.832, 270.419, 492.889],
    'seq06_11': [756.026, 756.832, 270.419, 492.889],
    'seq06_12': [758.792, 760.052, 266.208, 485.694],
    'seq07_01': [760.420, 761.683, 268.748, 486.778],
    'seq07_02': [760.420, 761.683, 268.747, 486.778],
    'seq07_03': [760.420, 761.683, 268.748, 486.778],
    'seq07_04': [760.420, 761.683, 268.748, 486.778],
    'seq07_05': [759.611, 760.513, 268.301, 491.445],
    'seq07_06': [759.611, 760.513, 268.301, 491.445],
    'seq07_07': [759.507, 760.646, 268.339, 487.215],
    'seq07_08': [759.507, 760.646, 268.339, 487.215],
    'seq08_01': [759.507, 760.646, 268.339, 487.215],
    'seq08_02': [757.958, 759.273, 263.450, 490.819],
    'seq08_03': [759.507, 760.646, 268.339, 487.215],
    'seq08_04': [758.754, 759.788, 269.799, 489.091],
    'seq08_05': [759.507, 760.646, 268.339, 487.215],
    'seq09_01': [758.792, 760.052, 266.208, 485.694],
    'seq09_02': [758.792, 760.051, 266.208, 485.693],
    'seq09_03': [758.792, 760.052, 266.208, 485.694],
    'seq09_04': [758.792, 760.052, 266.208, 485.694],
    'seq09_05': [758.792, 760.052, 266.208, 485.694],
    'seq10_01': [756.026, 756.832, 270.419, 492.889],
    'seq10_02': [756.026, 756.832, 270.418, 492.889],
    'seq10_03': [756.026, 756.832, 270.419, 492.889],
    'seq10_04': [756.026, 756.832, 270.419, 492.889],
    'seq10_05': [756.026, 756.832, 270.419, 492.889],
    'seq10_06': [755.581, 756.525, 269.186, 488.945],
    'seq10_07': [756.026, 756.832, 270.419, 492.889],
    'seq10_08': [759.611, 760.513, 268.301, 491.445]
}


def _import_RIO10_sequence(sequence_root: str, sequence_list: List[str],
                           kapture_dir_path: str, images_import_method: TransferAction,
                           force_overwrite_existing: bool):
    os.makedirs(kapture_dir_path, exist_ok=True)
    delete_existing_kapture_files(kapture_dir_path, force_erase=force_overwrite_existing)

    snapshots = kapture.RecordsCamera()
    depth_maps = kapture.RecordsDepth()
    trajectories = kapture.Trajectories()
    rigs = kapture.Rigs()
    sensors = kapture.Sensors()
    RIO10_seq_filename_re = re.compile(r'^frame-(?P<frame_id>\d{6})\.(?P<suffix>.*)\.(?P<ext>\w*)$')

    current_timestamp = 0
    for senquence_path in sequence_list:
        rgb_sensor_id = senquence_path + '_rgb'
        depth_sensor_id = senquence_path + '_depth'
        rig_sensor_id = senquence_path
        sensors[rgb_sensor_id] = kapture.Camera(CAMERA_TYPE, [WIDTH, HEIGHT] + INTRINSICS[senquence_path])
        sensors[depth_sensor_id] = kapture.Camera(CAMERA_TYPE, [WIDTH, HEIGHT] + INTRINSICS[senquence_path],
                                                  sensor_type=SENSOR_TYPE_DEPTH_CAM)
        rigs[rig_sensor_id, rgb_sensor_id] = kapture.PoseTransform()
        rigs[rig_sensor_id, depth_sensor_id] = kapture.PoseTransform()

        file_list = os.listdir(path.join(sequence_root, senquence_path))
        logger.debug(f'populating {senquence_path} files ...')
        RIO10_seq_filenames = {filename: RIO10_seq_filename_re.search(filename).groupdict()
                               for filename in sorted(file_list)
                               if RIO10_seq_filename_re.search(filename)}
        # reorg as shot[seq, id] = {color: , depth: , pose: , ...}
        shots = {}
        for timestamp, (filename, file_attribs) in enumerate(RIO10_seq_filenames.items()):
            shot_id = int(file_attribs['frame_id'])
            shots.setdefault(shot_id, {})[file_attribs['suffix']] = filename
        for timestamp, shot_id in enumerate(shots):
            shots[shot_id]['timestamp'] = shot_id + current_timestamp
        for shot in shots.values():
            snapshots[shot['timestamp'], rgb_sensor_id] = senquence_path + '/' + shot[RGB_SUFFIX]
            # kapture depth files are not png
            kapture_depth_map_filename = senquence_path + '/' + shot[DEPTH_SUFFIX][:-len('.png')]
            depth_maps[shot['timestamp'], depth_sensor_id] = kapture_depth_map_filename

            if POSE_SUFFIX in shot:
                pose_filepath = path.join(sequence_root, senquence_path, shot['pose'])
                pose_mat = np.loadtxt(pose_filepath)  # camera-to-world, 4×4 matrix in homogeneous coordinates
                with open(pose_filepath, 'r') as file:
                    if 'INF' in file.read():
                        timestamp = shot['timestamp']
                        image_name = shot[RGB_SUFFIX]
                        logger.debug(f'ts={timestamp}, name={image_name}: ignored inf pose')
                        continue
                rotation_mat = pose_mat[0:3, 0:3]
                position_vec = pose_mat[0:3, 3]
                rotation_quat = quaternion.from_rotation_matrix(rotation_mat)
                pose_world_from_cam = kapture.PoseTransform(r=rotation_quat, t=position_vec)
                pose_cam_from_world = pose_world_from_cam.inverse()
                trajectories[shot['timestamp'], rig_sensor_id] = pose_cam_from_world
        current_timestamp += max(shots.keys()) + 1 + TIMESTAMP_SEQUENCE_BREAK

    # import (copy) image files.
    logger.info('copying image files ...')
    image_filenames = snapshots.data_list()
    import_record_data_from_dir_auto(sequence_root, kapture_dir_path, image_filenames, images_import_method)

    # import (copy) depth map files.
    logger.info('converting depth files ...')
    depth_map_filenames = kapture.io.records.records_to_filepaths(depth_maps, kapture_dir_path)
    hide_progress = logger.getEffectiveLevel() > logging.INFO
    for depth_map_filename, depth_map_filepath_kapture in tqdm(depth_map_filenames.items(), disable=hide_progress):
        depth_map_filepath_rio10 = path.join(sequence_root, depth_map_filename + '.png')
        depth_map = np.array(Image.open(depth_map_filepath_rio10))
        # depth maps is in mm in rio10, convert it to meters
        depth_map = depth_map.astype(np.float32) * DEPTH_TO_METER
        kapture.io.records.depth_map_to_file(depth_map_filepath_kapture, depth_map)

    imported_kapture = kapture.Kapture(
        records_camera=snapshots,
        records_depth=depth_maps,
        rigs=rigs,
        trajectories=trajectories or None,
        sensors=sensors)
    logger.info('writing imported data ...')
    kapture_to_dir(kapture_dir_path, imported_kapture)


def import_RIO10(dRIO10_path: str,
                 kapture_dir_path: str,
                 force_overwrite_existing: bool = False,
                 images_import_method: TransferAction = TransferAction.skip
                 ) -> None:
    """
    Imports RIO10 dataset and save them as kapture.

    :param dRIO10_path: path to the RIO10 root path (contains folders named sceneXX)
    :param kapture_dir_path: path to kapture output top directory
    :param force_overwrite_existing: Silently overwrite kapture files if already exists.
    :param images_import_method: choose how to import actual image files
    """
    os.makedirs(kapture_dir_path, exist_ok=True)

    scenes = list(range(1, 11))
    for scene_id in scenes:
        logger.info(f'import scene {scene_id}...')
        scene_path = path.join(dRIO10_path, f'scene{scene_id:02d}')
        if not path.isdir(scene_path):
            logger.warning(f'{scene_path} does not exist')
            continue
        seq_path = path.join(scene_path, f'seq{scene_id:02d}')
        if not path.isdir(seq_path):
            logger.warning(f'{seq_path} does not exist')
            continue
        seq_list = os.listdir(seq_path)

        seq_subfolder_re = re.compile(r'seq(?P<scene_id>\d{2})_(?P<seq_id>\d{2})')

        seq_subfolders = {}
        for seq_subfolder in seq_list:
            seq_subfolder_search = seq_subfolder_re.search(seq_subfolder)
            if seq_subfolder_search:
                seq_subfolder_re_dict = seq_subfolder_search.groupdict()
                if int(seq_subfolder_re_dict['scene_id']) != scene_id:
                    logger.warning(f'{seq_path} : found a sequence with wrong sequence id, skipped')
                    continue
                seq_subfolders[int(seq_subfolder_re_dict['seq_id'])] = seq_subfolder
        if MAPPING_SEQUENCE_ID in seq_subfolders:
            # 1 is always the mapping sequence
            _import_RIO10_sequence(seq_path, [seq_subfolders[MAPPING_SEQUENCE_ID]],
                                   path.join(kapture_dir_path, f'scene{scene_id:02d}', 'mapping'),
                                   images_import_method, force_overwrite_existing)
        if VALIDATION_SEQUENCE_ID in seq_subfolders:
            # 2 is always the mapping sequence
            _import_RIO10_sequence(seq_path, [seq_subfolders[VALIDATION_SEQUENCE_ID]],
                                   path.join(kapture_dir_path, f'scene{scene_id:02d}', 'validation'),
                                   images_import_method, force_overwrite_existing)
        testing_sequences = [v
                             for k, v in sorted(seq_subfolders.items())
                             if k != MAPPING_SEQUENCE_ID and k != VALIDATION_SEQUENCE_ID]
        if len(testing_sequences) > 0:
            _import_RIO10_sequence(seq_path, testing_sequences,
                                   path.join(kapture_dir_path, f'scene{scene_id:02d}', 'testing'),
                                   images_import_method, force_overwrite_existing)


def import_RIO10_command_line() -> None:
    """
    Imports RGB-D Dataset RIO10 and save them as kapture using the parameters given on the command line.
    """
    parser = argparse.ArgumentParser(
        description='Imports RGB-D Dataset RIO10 files to the kapture format.')
    parser_verbosity = parser.add_mutually_exclusive_group()
    parser_verbosity.add_argument(
        '-v', '--verbose', nargs='?', default=logging.WARNING, const=logging.INFO,
        action=kapture.utils.logging.VerbosityParser,
        help='verbosity level (debug, info, warning, critical, ... or int value) [warning]')
    parser_verbosity.add_argument(
        '-q', '--silent', '--quiet', action='store_const', dest='verbose', const=logging.CRITICAL)
    parser.add_argument('-f', '-y', '--force', action='store_true', default=False,
                        help='Force delete output if already exists.')
    # import ###########################################################################################################
    parser.add_argument('-i', '--input', required=True, help='input path to RIO10 root path')
    parser.add_argument('--image_transfer', type=TransferAction, default=TransferAction.link_absolute,
                        help=f'How to import images [link_absolute], '
                             f'choose among: {", ".join(a.name for a in TransferAction)}')
    parser.add_argument('-o', '--output', required=True, help='output directory.')
    ####################################################################################################################
    args = parser.parse_args()

    logger.setLevel(args.verbose)
    if args.verbose <= logging.DEBUG:
        # also let kapture express its logs
        kapture.utils.logging.getLogger().setLevel(args.verbose)

    import_RIO10(dRIO10_path=args.input,
                 kapture_dir_path=args.output,
                 force_overwrite_existing=args.force,
                 images_import_method=args.image_transfer)


if __name__ == '__main__':
    import_RIO10_command_line()



#!/usr/bin/env python3
# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Script to print statistics about kapture data.
"""

import argparse
import logging
import os
import os.path as path
import re
from datetime import datetime

import path_to_kapture  # noqa: F401
import kapture
import kapture.utils.computation as computation
import kapture.utils.logging
import kapture.io.csv

logger = logging.getLogger('kapture_data_date')

KAPTURE_FILE_PARSING_RE = re.compile('.*(records|trajectories).*\\.txt')


def format_timestamp_range(first_timestamp: int, last_timestamp: int) -> str:
    """
    Format a time range with unit.

    :param first_timestamp: first timestamp
    :param last_timestamp: last timestamp
    :return string: formatted time range
    """
    timestamp_len = 1
    try:
        timestamp_len = computation.num_digits(first_timestamp)
        delta_len = 10 - timestamp_len  # datetime expects epoch timestamp in seconds
        factor = 10**delta_len
        dts = [datetime.utcfromtimestamp(first_timestamp * factor),
               datetime.utcfromtimestamp(last_timestamp * factor)]
        timestamp_format_d = {
            'date': '%Y/%m/%d',
            'time': '%H:%M:%S.%f'
        }
        timestamp_parts = [
            {
                pname: dt.strftime(pformat)
                for pname, pformat in timestamp_format_d.items()
            }
            for dt in dts
        ]
        timestamp_str = timestamp_parts[0]['date'] + ' ' + timestamp_parts[0]['time']
        timestamp_str += ' -> '
        if timestamp_parts[0]['date'] != timestamp_parts[1]['date']:
            timestamp_str = timestamp_parts[0]['date'] + ' '
        timestamp_str += timestamp_parts[1]['time'] + ' GMT'
        return timestamp_str
    except ValueError as _:  # noqa: F841
        return f'{first_timestamp} -> {last_timestamp}' \
               + f' ** FAIL to parse as posix timestamp of {timestamp_len} digits'


def print_info(kapture_path: str, kapture_filename: str) -> None:
    """
    Prints some info on given file that should contains records with timestamps

    :param kapture_path: kapture top directory
    :param kapture_filename: full path of a valid kapture file
    """
    # Check version to test if this is a real kapture file
    kapture_file_path = path.join(kapture_path, kapture_filename)
    version = kapture.io.csv.get_version_from_csv_file(kapture_file_path)
    if version is None:
        logger.debug(f'{kapture_filename} not a kapture file')
    else:
        # Read the file
        last_data_line = ''
        nb_lines = 0
        with open(kapture_file_path) as f:
            # Count lines
            for _ in f:
                nb_lines += 1
            # Reset to read some important lines
            f.seek(0, os.SEEK_SET)
            # Skip header
            f.readline()
            f.readline()
            first_data_line = f.readline()
            if first_data_line:
                last_data_line = kapture.io.csv.get_last_line(f)
        first_timestamp = 0
        last_timestamp = 0
        # data line are comma separated lines with the timestamp as first value
        try:
            if first_data_line:
                first_timestamp = int(first_data_line.split(',')[0])
            if last_data_line:
                last_timestamp = int(last_data_line.split(',')[0])
        except ValueError:
            pass
        if first_timestamp > 0 and last_timestamp > 0:
            timestamp_range_str = format_timestamp_range(first_timestamp, last_timestamp)
            timestamp_len1 = computation.num_digits(first_timestamp)
            timestamp_len2 = computation.num_digits(last_timestamp)
            timestamp_len_str = f'{timestamp_len1}' if timestamp_len1 == timestamp_len2\
                else f'{timestamp_len1}-{timestamp_len2}'
            print(f'{kapture_filename:42s} timestamp {timestamp_len_str} digits from {timestamp_range_str}'
                  f' : {(nb_lines-2):12,d} records'.replace(',', ' '))


def do_print(kapture_path: str) -> None:
    """
    Print out kapture data:

    :param kapture_path: full path to kapture directory.
    """
    dirs_to_examine = set()
    for k_path in kapture.io.csv.CSV_FILENAMES.values():
        dirs_to_examine.add(path.dirname(k_path))
    # Look for files which might contain kapture data
    for k_dir in sorted(dirs_to_examine):
        kapture_dir = path.join(kapture_path, k_dir)
        if path.exists(kapture_dir):
            if path.isdir(kapture_dir):
                # print
                logger.debug(f'browsing directory {k_dir}')
                # Search for records or trajectory like files
                kapture_files = []
                for entry in os.listdir(kapture_dir):
                    if path.isfile(path.join(kapture_dir, entry)):
                        if KAPTURE_FILE_PARSING_RE.match(entry) is not None:
                            kapture_files.append(entry)
                for kapture_file in sorted(kapture_files):
                    print_info(kapture_path, path.join(k_dir, kapture_file))
            else:
                logger.fatal(f'{kapture_dir} is not a directory')
        else:
            logger.debug(f'{kapture_dir} does not exist')


def print_command_line() -> None:
    """
    Do the print using the parameters given on the command line.
    """

    parser = argparse.ArgumentParser(description='Print statistics about kapture records files with timestamps.')
    parser_verbosity = parser.add_mutually_exclusive_group()
    parser_verbosity.add_argument(
        '-v', '--verbose', nargs='?', default=logging.WARNING, const=logging.INFO,
        action=kapture.utils.logging.VerbosityParser,
        help='verbosity level (debug, info, warning, critical, ... or int value) [warning]')
    parser_verbosity.add_argument(
        '-q', '--silent', '--quiet', action='store_const', dest='verbose', const=logging.CRITICAL)
    parser.add_argument('-k', '--kapture', required=True,
                        help='path to kapture data root directory.')
    args = parser.parse_args()

    logger.setLevel(args.verbose)
    if args.verbose <= logging.DEBUG:
        # also let kapture express its logs
        kapture.utils.logging.getLogger().setLevel(args.verbose)

    do_print(path.abspath(args.kapture))


if __name__ == '__main__':
    print_command_line()



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Extract GNSS data from NMEA 0183 files
"""
from datetime import datetime, timedelta
import logging
import pytz
import re
from typing import Tuple, Optional

import kapture
from kapture.io.csv import table_from_file

logger = logging.getLogger('nmea')


def day_from_nmea_str(date_str: str) -> datetime:
    return datetime.strptime(date_str, '%d%m%y')


time_parser = re.compile(r'^(?P<hours>\d{2})(?P<minutes>\d{2})(?P<seconds>\d{2})\.(?P<milliseconds>\d+)$')


def nmea_str_to_datetime(
        day: datetime,
        time_str: str) -> datetime:
    """

    :param day: current date (assuming time part is set to 00:00:00)
    :param time_str: a string representing the nmea time (eg 124638.20 for 12h 46m 38s 200ms)
    :return: a datetime object with timestamp and timezone set.
    """
    time_params = time_parser.match(time_str)
    assert time_params is not None
    assert day.hour == 0
    time_params = time_params.groupdict()
    if len(time_params['milliseconds']) < 3:  # make sure milliseconds are on 3 digits
        time_params['milliseconds'] += '0' * (3 - len(time_params['milliseconds']))
    time_params = {k: int(v) for k, v in time_params.items()}
    time_of_day = timedelta(**time_params)
    timestamp = day + time_of_day
    timestamp = pytz.UTC.localize(timestamp)  # timestamp is guaranteed UTC
    return timestamp


def nmea_coord_to_lla(
    lat_str: str, lat_cardinal: str,
    lon_str: str, lon_cardinal: str,
    alt_str: str
) -> Tuple[float, float, float]:
    """
    converts the nmea quintuplet (4811.7605691,N,01137.1631304,E,494.6412) to decimal gps coordinates
    :param lat_str: latitude in nmea format
    :param lat_cardinal: latitude cardinal (N for north, S for south)
    :param lon_str: longitude in nmea format
    :param lon_cardinal: logitude cardinal (E for east, W for west)
    :param alt_str: altitude in nmea format
    :return:
    """
    lat_deg, lat_min = float(lat_str[0:2]), float(lat_str[2:])
    lat_dec = lat_deg + lat_min / 60.0
    lon_split = lon_str.index('.') - 2  # degree part may be 2 or 3 digits long
    lon_deg, lon_min = float(lon_str[0:lon_split]), float(lon_str[lon_split:])
    lon_dec = lon_deg + lon_min / 60.0
    alt_dec = float(alt_str)
    if 'S' == lat_cardinal:
        lat_dec *= -1
    if 'W' == lon_cardinal:
        lon_dec *= -1

    return lat_dec, lon_dec, alt_dec


def extract_gnss_from_nmea(
        nmea_file_path: str,
        gnss_id: Optional[str]='gnss'
):
    """
    Extract coordinates from NMEA file, returns sensors with the new gnss sensor in it and gnss records.
    - Gnss timestamps are: in nanoseconds

    :param nmea_file_path: path to nmea input file..
    :param gnss_id: ID of the GNSS sensor.
    :return:
    """
    # only load sensors + records_data:

    # add new gps ids to sensors
    gnss_kapture_sensors = kapture.Sensors()
    gnss_kapture_sensors[gnss_id] = kapture.Sensor(
        sensor_type='gnss', name=gnss_id,
        sensor_params=['EPSG:4326'])  # aka WGS84
    #  x, y, z, utc, dop
    records_gnss = kapture.RecordsGnss()
    day = None

    with open(nmea_file_path, 'rt') as file:
        table = table_from_file(file)
        for trame in table:
            trame_type = trame[0]
            if '$GPRMC' == trame_type:  # retrieve date only (dont bother with coords, no altitude available)
                time_str, date_str = trame[1], trame[9]
                day = day_from_nmea_str(date_str)

            if day is None:  # dont bother do the rest if date is not retrieved yet.
                continue
            if '$GPGGA' == trame_type:  # GPS coordinates
                # UTC time, Latitude, hemisphere, Longitude, meridian, 1 (for GPS), #satellites used, HDOP, Altitude,...
                time_str, lat, ns, lon, ew, t, sats, dop, alt = trame[1:10]
                try:
                    timestamp_dt = nmea_str_to_datetime(day, time_str)
                    timestamp_ns = int(timestamp_dt.timestamp() * 1e9)
                    lat, lon, alt = nmea_coord_to_lla(lat, ns, lon, ew, alt)
                    # x, y, z, utc, dop
                    gps_record = kapture.RecordGnss(x=lon, y=lat, z=alt, utc=timestamp_ns, dop=dop)
                    records_gnss[timestamp_ns, gnss_id] = gps_record
                except Exception as e:
                    logger.debug(f'skipping malformed nmea frame {e}')

    return gnss_kapture_sensors, records_gnss



#!/usr/bin/env python3
# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Convert kapture data to ETH-Microsoft Long-term Visual Localization challenge format
See https://github.com/cvg/visloc-iccv2021

This line should store the
result as `timestamp/camera_or_rig_id qw qx qy qz tx ty tz`.
`qw qx qy qz` represents the **rotation** from world to camera coordinates as a
**unit quaternion**. `tx ty tz` is the camera **translation** (**not the camera position**).

"""
import logging
import os
import pathlib
import argparse
from tqdm import tqdm

import path_to_kapture  # noqa: F401
import kapture
from kapture.core.Trajectories import rigs_recover
import kapture.utils.logging
import kapture.io.features
import kapture.io.csv

logger = logging.getLogger('ETH_MS_LTVL')


def export_ETH_MS_LTVL(kapture_dirpath: str,
                       ltvl_filepath: str) -> None:
    """
    Export kapture data to ETH-Microsoft Long-term Visual Localization challenge format file.

    :param kapture_dirpath: kapture data top directory
    :param ltvl_filepath: LTVL file path to write

    """
    skip_heavy = [kapture.RecordsLidar, kapture.RecordsWifi,
                  kapture.Keypoints, kapture.Descriptors,
                  kapture.GlobalFeatures, kapture.Matches,
                  kapture.Points3d, kapture.Observations]
    kapture_data = kapture.io.csv.kapture_from_dir(kapture_dirpath, skip_list=skip_heavy)

    # no nested rigs in ETH-MS
    assert kapture_data.rigs is not None
    assert kapture_data.trajectories is not None
    trajectories = rigs_recover(kapture_data.trajectories, kapture_data.rigs)

    # write the file

    p = pathlib.Path(ltvl_filepath)
    os.makedirs(str(p.parent.resolve()), exist_ok=True)
    with open(ltvl_filepath, 'wt') as f:
        for timestamp, sensor_id, pose in tqdm(kapture.flatten(trajectories),
                                               disable=logger.getEffectiveLevel() > logging.INFO):
            line = [str(timestamp) + '/' + sensor_id] + pose.r_raw + pose.t_raw
            line = ' '.join(str(v) for v in line) + '\n'
            f.write(line)


def export_ETH_MS_LTVL_command_line() -> None:
    """
    Do the ETH-Microsoft LTVL export using the parameters given on the command line.
    """
    parser = argparse.ArgumentParser(
        description=('convert file to ETH-Microsoft Long-term Visual Localization challenge format '
                     '(https://www.visuallocalization.net/submission/).'))
    parser_verbosity = parser.add_mutually_exclusive_group()
    parser_verbosity.add_argument(
        '-v', '--verbose', nargs='?', default=logging.WARNING, const=logging.INFO,
        action=kapture.utils.logging.VerbosityParser,
        help='verbosity level (debug, info, warning, critical, ... or int value) [warning]')
    parser_verbosity.add_argument(
        '-q', '--silent', '--quiet', action='store_const', dest='verbose', const=logging.CRITICAL)
    parser.add_argument('-f', '-y', '--force', action='store_true', default=False,
                        help='Force delete output if already exists.')
    # export ###########################################################################################################
    parser.add_argument('-i', '--input', required=True, help='input path to kapture directory')
    parser.add_argument('-o', '--output', required=True, help='output file.')

    ####################################################################################################################
    args = parser.parse_args()

    logger.setLevel(args.verbose)
    if args.verbose <= logging.DEBUG:
        # also let kapture express its logs
        kapture.utils.logging.getLogger().setLevel(args.verbose)

    export_ETH_MS_LTVL(args.input, args.output)


if __name__ == '__main__':
    export_ETH_MS_LTVL_command_line()



#!/usr/bin/env python3
# Copyright 2021-present NAVER Corp. Under BSD 3-clause license

"""
This script imports a trajectory file with images in the kapture format
The trajectory file contains the following info on every line: timestamp, camera id, image path, 6d pose

"""

import argparse
import logging
import os
import os.path as path
import PIL
from PIL import Image
import quaternion
import numpy as np
from typing import List

# kapture
import path_to_kapture  # noqa: F401
import kapture
import kapture.utils.logging
from kapture.io.csv import table_from_file, kapture_to_dir
from kapture.io.records import TransferAction, import_record_data_from_dir_auto
from kapture.io.structure import delete_existing_kapture_files
from kapture.utils.open_cv import import_opencv_camera_calibration

logger = logging.getLogger('import_image_list_with_poses')


def import_image_list_with_poses(images_list_file_path: str,
                                 images_dir_path: str,
                                 camera_calibration_file_path: str,
                                 kapture_path: str,
                                 force_overwrite_existing: bool = False,
                                 images_import_method: TransferAction = TransferAction.skip,
                                 do_not_import_images: bool = False) -> None:
    """
    Imports the list of images with their poses to a kapture. This creates only images and cameras.

    :param images_list_file_path: file containing the list of images with their poses
    :param images_dir_path: top directory of the images path. If not defined, the images path in the list file
        must be full path.
    :param camera_calibration_file_path: path to the calibration camera file in opencv format.
    If missing, will compute a simple camera model based on the images
    :param kapture_path: path to kapture root directory.
    :param force_overwrite_existing: Silently overwrite kapture files if already exists.
    :param images_import_method: choose how to import actual image files.
    :param do_not_import_images: when true, do not import the image files.
    """
    if not path.exists(images_list_file_path):
        raise FileNotFoundError(images_list_file_path)
    os.makedirs(kapture_path, exist_ok=True)
    delete_existing_kapture_files(kapture_path, force_erase=force_overwrite_existing)

    cameras = kapture.Sensors()
    camera_model_provided = False
    if camera_calibration_file_path:
        if not path.exists(camera_calibration_file_path):
            raise FileNotFoundError(camera_calibration_file_path)
        camera = import_opencv_camera_calibration(camera_calibration_file_path)
        cameras[camera.name] = camera
        camera_model_provided = True
    images = kapture.RecordsCamera()
    trajectories = kapture.Trajectories()
    filename_list: List[str] = []
    image_name_list: List[str] = []

    logger.info(f'Loading data from {images_list_file_path}')
    # Find the top directory of all image files
    with open(images_list_file_path, 'rt') as file:
        images_list_data = table_from_file(file)
        for image_line in images_list_data:
            image_path = image_line[2]
            if not do_not_import_images and not images_dir_path and not path.isabs(image_path):
                raise ValueError(f'Can not import a relative image file without top directory: {image_path}')
            if not path.isabs(image_path) and images_dir_path:
                image_path = path.join(images_dir_path, image_path)
            filename_list.append(image_path)
    common_images_path = path.commonpath(filename_list)

    # Parse image list data file
    with open(images_list_file_path, 'rt') as file:
        images_list_data = table_from_file(file)
        for image_line in images_list_data:
            timestamp, camera_id, image_path, qw, qx, qy, qz, tx, ty, tz = image_line
            # Add top directory if necessary
            if not path.isabs(image_path) and images_dir_path:
                image_path = path.join(images_dir_path, image_path)
            if path.exists(image_path):
                if not camera_model_provided:
                    # Create corresponding camera model
                    model = kapture.CameraType.UNKNOWN_CAMERA.value
                    try:
                        # lazy open
                        with Image.open(image_path) as im:
                            width, height = im.size
                            model_params = [width, height]
                            cameras[camera_id] = kapture.Camera(model, model_params, camera_id)
                    except (OSError, PIL.UnidentifiedImageError):
                        # It is not a valid image: skip it
                        logger.info(f'Skipping invalid image file {image_path}')
                if not do_not_import_images:
                    image_name = path.relpath(image_path, common_images_path)
                    image_name_list.append(image_name)
                    images[(int(timestamp), camera_id)] = image_name
            else:
                logger.debug(f'Missing image file {image_path}')
            # Create pose
            rotation = None
            translation = None
            if qw != '' and qx != '' and qy != '' and qz != '':
                rotation = quaternion.from_float_array([float(qw), float(qx), float(qy), float(qz)])
            if tx != '' and ty != '' and tz != '':
                translation = np.array([[float(tx)], [float(ty)], [float(tz)]], dtype=np.float)
            pose = kapture.PoseTransform(rotation, translation)
            trajectories[(int(timestamp), camera_id)] = pose

    # import (copy) image files.
    if not do_not_import_images:
        logger.info('import image files ...')
        import_record_data_from_dir_auto(common_images_path, kapture_path, image_name_list, images_import_method)

    # pack into kapture format
    imported_kapture = kapture.Kapture(sensors=cameras, records_camera=images, trajectories=trajectories)
    logger.info('writing imported data...')
    kapture_to_dir(kapture_path, imported_kapture)


def import_image_list_with_poses_command_line() -> None:
    """
    Do the images list import to kapture using the parameters given on the command line.
    """
    parser = argparse.ArgumentParser(description='import images with their poses (trajectory+images)'
                                                 ' to the kapture format.')
    parser_verbosity = parser.add_mutually_exclusive_group()
    parser_verbosity.add_argument(
        '-v', '--verbose', nargs='?', default=logging.WARNING, const=logging.INFO,
        action=kapture.utils.logging.VerbosityParser,
        help='verbosity level (debug, info, warning, critical, ... or int value) [warning]')
    parser_verbosity.add_argument(
        '-q', '--silent', '--quiet', action='store_const', dest='verbose', const=logging.CRITICAL)
    parser.add_argument('-f', '--force', action='store_true', default=False,
                        help='force delete output if already exists.')
    # import ###########################################################################################################
    parser.add_argument('-l', '--images_list_file', required=True,
                        help='path to images (with their poses) list file')
    parser.add_argument('--top_dir_images', default='',
                        help='path to top directory of images.'
                             ' Necessary only if the images path are relative in the list file')
    parser.add_argument('-c', '--camera_model_file',
                        help='path to opencv camera calibration YAML file. Must be in YAML 1.1.'
                             'If missing, will make a model based on images')
    parser.add_argument('-k', '--kapture', required=True, help='kapture directory')
    parser.add_argument('--image_action', type=TransferAction, default=TransferAction.copy,
                        help=f'How to import images [copy], '
                             f'choose among: {", ".join(a.name for a in TransferAction)}')
    parser.add_argument('--do_not_import_images', action='store_true', default=False,
                        help='Do not import the images files, but their poses (the trajectory) only.')
    ####################################################################################################################
    args = parser.parse_args()

    logger.setLevel(args.verbose)
    if args.verbose <= logging.DEBUG:
        # also let kapture express its logs
        kapture.utils.logging.getLogger().setLevel(args.verbose)

    import_image_list_with_poses(args.images_list_file, args.top_dir_images, args.camera_model_file,
                                 args.kapture, args.force,
                                 args.image_action, args.do_not_import_images)


if __name__ == '__main__':
    import_image_list_with_poses_command_line()



# Copyright 2021-present NAVER Corp. Under BSD 3-clause license

"""
OpenCV related helper operations.
"""

import os.path as path
import pathlib
import yaml

from kapture.core.Sensors import CameraType, Camera


def import_opencv_camera_calibration(camera_info_file_path: str) -> Camera:
    """
    Parse an OpenCV camera calibration file to insert in the sensor list.

    :param camera_info_file_path: path to the camera info file
    :return: the camera name and its parameters
    """
    camera_info_file_path = path.abspath(camera_info_file_path)
    with open(camera_info_file_path) as file:
        # First check a glitch coming from OpenCV 3: the YAML file produced is not compliant with PYYAML
        first_line = file.readline()
        if first_line.startswith('%YAML:'):
            first_line = ''
        yaml_content = first_line
        for line in file.readlines():
            # Check if they are some unsupported !!
            start_invalid = line.find('!!')
            if start_invalid != -1:
                # Remove everything after the !! except the end of line character
                line = line[0:start_invalid] + line[-1:]
            yaml_content = yaml_content + line
        camera_info = yaml.full_load(yaml_content)
        if 'camera_name' in camera_info:
            camera_name = camera_info['camera_name']
        else:
            # Comes straight from OpenCV without name: use file name instead
            camera_name = pathlib.Path(path.basename(camera_info_file_path)).stem
        w, h = camera_info['image_width'], camera_info['image_height']
        camera_matrix = camera_info['camera_matrix']['data']
        fx, cx, fy, cy = camera_matrix[0], camera_matrix[2], camera_matrix[4], camera_matrix[5]
        # Will use opencv model with parameters: w, h, fx, fy, cx, cy, k1, k2, p1, p2
        distortion_coefficients = camera_info['distortion_coefficients']['data']
        camera_type = CameraType.OPENCV
        if distortion_coefficients[4] == float(0.0):
            # No k3: opencv model with k1 k2 p1 p2
            camera_parameters = [w, h, fx, fy, cx, cy] + distortion_coefficients[0:4]
        else:
            camera_type = CameraType.FULL_OPENCV
            # must provide k1, k2, p1, p2, k3, k4, k5, k6
            # k3 is here but not necessarily k4, k5 and k6
            k4 = distortion_coefficients[5] if len(distortion_coefficients) > 5 else float(0)
            k5 = distortion_coefficients[6] if len(distortion_coefficients) > 6 else float(0)
            k6 = distortion_coefficients[7] if len(distortion_coefficients) > 7 else float(0)
            camera_parameters = [w, h, fx, fy, cx, cy] + distortion_coefficients[0:5] + [k4, k5, k6]
        camera = Camera(camera_type, camera_parameters, camera_name)
        return camera



#!/usr/bin/env python3
# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Script to import a 4seasons model into a kapture.
The data structure is defined here:
    https://www.4seasons-dataset.com/documentation

The 4Seasons dataset contains recordings from a stereo-inertial camera system coupled with a high-end RTK-GNSS.

For each sequence, the recorded data is stored in the following structure:
├── KeyFrameData
├── distorted_images ...
├── undistorted_images
│   ├── cam0
│   └── cam1
├── GNSSPoses.txt
├── Transformations.txt
├── imu.txt
├── result.txt
├── septentrio.nmea
└── times.txt

The calibration folder has the following structure:
├── calib_0.txt
├── calib_1.txt
├── calib_stereo.txt
├── camchain.yaml
├── undistorted_calib_0.txt
├── undistorted_calib_1.txt
└── undistorted_calib_stereo.txt
"""

import argparse
import logging
import os
import os.path as path
import re
import math
import numpy as np
import quaternion
from glob import glob
from tqdm import tqdm
from dataclasses import dataclass
from typing import List, Dict, Tuple, Union
# kapture
import path_to_kapture  # noqa: F401
import kapture
import kapture.utils.logging
from kapture.io.structure import delete_existing_kapture_files
from kapture.io.csv import kapture_to_dir
import kapture.io.features
from kapture.io.records import TransferAction, import_record_data_from_dir_auto
from kapture.utils.logging import getLogger
from kapture.converter.nmea.import_nmea import extract_gnss_from_nmea

logger = getLogger()

MASTER_CAM_ID = 'cam0'
RIG_ID = 'car'
ACCELEROMETER_ID = 'accelero'
GYROSCOPE_ID = 'gyro'
DEPTH_ID = 'depth_sensor'

q = quaternion.from_rotation_matrix(np.array([[-1, 0, 0], [0, 1, 0], [0, 0, -1]]))
CAM_AXES_KAPTURE_FROM_4SEASONS = kapture.PoseTransform(r=q)
CAM_AXES_4SEASONS_FROM_KAPTURE = CAM_AXES_KAPTURE_FROM_4SEASONS.inverse()


def load_4seasons_cameras(
        calibration_dir_path: str
) -> (kapture.Sensors, kapture.Rigs):
    # this dataset is made with a single stereo camera (2 cams).
    sensors = kapture.Sensors()
    """
    Pinhole 501.4757919305817 501.4757919305817 421.7953735163109 167.65799492501083 0.0 0.0 0.0 0.0
    800 400
    crop
    800 400
    """
    intrinsic_file_names = {
        'cam0': 'undistorted_calib_0.txt',
        'cam1': 'undistorted_calib_1.txt',
    }
    cam_names = {
        'cam0': 'cam_left',
        'cam1': 'cam_right'
    }
    for cam_id, intrinsic_file_name in intrinsic_file_names.items():
        intrinsic_file_path = path.join(calibration_dir_path, intrinsic_file_name)
        with open(intrinsic_file_path, 'r') as intrinsic_file:
            line = intrinsic_file.readline().split(' ')
            # 1st line looks like:
            #     Pinhole 501.4757919305817 501.4757919305817 421.7953735163109 167.65799492501083 0.0 0.0 0.0 0.0
            # assuming fx fy cx cy distortion coef (null)
            if line[0] != 'Pinhole':  # sanity check
                raise ValueError(f'unexpected camera model {line[0]} (only is Pinhole valid).')
            fx, fy, cx, cy = (float(e) for e in line[1:5])

            line = intrinsic_file.readline().split(' ')
            # second line looks like :
            #   800 400
            # assuming image_width image_height
            w, h = (int(e) for e in line)
            sensors[cam_id] = kapture.Camera(kapture.CameraType.PINHOLE, [w, h, fx, fy, cx, cy], name=cam_names[cam_id])

    # rigs
    rigs = kapture.Rigs()
    stereo_matrix_file_name = path.join(calibration_dir_path, "undistorted_calib_stereo.txt")
    # contains the 4x4 matrix denoting the rigid transformation from the right to the left camera.
    # cam0 = left, cam1 = right
    car_from_cam1_matrix = np.loadtxt(stereo_matrix_file_name)
    r = car_from_cam1_matrix[0:3, 0:3]
    q = quaternion.from_rotation_matrix(r)
    t = car_from_cam1_matrix[0:3, 3]
    cam0_4s_from_cam1_4s = kapture.PoseTransform(q, t)
    cam0_from_cam1 = kapture.PoseTransform.compose([CAM_AXES_KAPTURE_FROM_4SEASONS,
                                                    cam0_4s_from_cam1_4s,
                                                    CAM_AXES_4SEASONS_FROM_KAPTURE])
    rigs[RIG_ID, MASTER_CAM_ID] = kapture.PoseTransform()
    rigs[RIG_ID, 'cam1'] = cam0_from_cam1.inverse()

    # trajectories from
    return sensors, rigs


def load_times_ids(
        times_file_path: str
) -> Dict[int, int]:
    """ load times.txt file to a dict linking shot_id to timestamp in nanoseconds """
    table = np.loadtxt(times_file_path)
    shots_ids = table[:, 0].astype(int)
    timestamps_ns = (table[:, 1] * 1e9).astype(int)
    shot_id_to_timestamp = {}
    for shot_id, timestamp_ns in zip(shots_ids, timestamps_ns):
        timestamp_ns = int(timestamp_ns)
        shot_id = f'{shot_id:19d}'
        # shot_id should match timestamp in ns
        assert str(timestamp_ns) == shot_id
        shot_id_to_timestamp[shot_id] = timestamp_ns

    return shot_id_to_timestamp


def guess_timestamp(
        shot_id: Union[str, int]
) -> int:
    """
     make up a timestamp (in ns) from the shot_id (in s).
     Always prefer the given table from times.txt since last decimal digits may differ.
    """
    # 1602074877621449728 -> 1602074877.621449728 sec -> 1602074877621449.728 u sec
    if not isinstance(shot_id, int):
        shot_id = int(shot_id)
    timestamp = int(float(shot_id) * 1e-9 * 1e6)
    return timestamp


def import_4seasons_images(
        recording_dir_path: str,
        kapture_dir_path: str,
        shot_id_to_timestamp: Dict[str, int],
        sensors: kapture.Sensors,
        images_import_method: TransferAction,
) -> kapture.RecordsCamera:
    """
    imports and copy image files.

    :param recording_dir_path:
    :param kapture_dir_path:
    :param shot_id_to_timestamp:
    :param sensors:
    :param images_import_method:
    :return:
    """
    kapture_images = kapture.RecordsCamera()
    logger.info('importing images ...')
    season_image_dir_path = recording_dir_path
    for sensor_id in sensors:
        for shot_id, timestamp_ns in shot_id_to_timestamp.items():
            image_file_name = path.join('undistorted_images', sensor_id, f'{shot_id}.png')
            # check image file is available
            if not path.isfile(path.join(season_image_dir_path, image_file_name)):
                logger.warning(f'image file is missing (and ignored) : {image_file_name}')
                # just throw it away
                continue
            kapture_images[timestamp_ns, sensor_id] = image_file_name

    filename_list = [f for _, _, f in kapture.flatten(kapture_images)]
    import_record_data_from_dir_auto(
        source_record_dirpath=season_image_dir_path,
        destination_kapture_dirpath=kapture_dir_path,
        filename_list=filename_list,
        copy_strategy=images_import_method)
    return kapture_images


KEYFRAME_TAG_POSE = '# camToWorld:'
KEYFRAME_TAG_CAM = '# fx, fy, cx, cy, width, height'
KEYFRAME_TAG_DEPTH = '# color information'


def read_until_tag(file, tag: str) -> str:
    """
     reads given file until it finds the given tags. Helps parsing basic 4seasons text files.
    it returns the line with the tag.
    """
    while True:
        line = file.readline()
        if line == '':
            raise ValueError(f'{tag} not found in file.')
        if line.startswith(tag):
            return line


def load_4season_pose_from_keyframe_data(
        keyframes_file_path: str
) -> kapture.PoseTransform:
    """
    Imports pose from keyframe file. Note this pose is the result of VIO.

    :param keyframes_file_path:
    :return:
    """
    # read pose
    with open(keyframes_file_path, 'rt') as file:
        read_until_tag(file, KEYFRAME_TAG_POSE)
        # translation vector, rotation quaternion
        line = file.readline()
    line = [float(v) for v in line.split(',')]
    tx, ty, tz, qx, qy, qz, qw = line  # NOTE: qw is at the end
    car_from_world = kapture.PoseTransform(r=[qw, qx, qy, qz], t=[tx, ty, tz]).inverse()
    # pose found
    return car_from_world


@dataclass
class DepthData:
    f: Tuple[float, float]
    c: Tuple[float, float]
    image_size: Tuple[int, int]
    coords: np.ndarray


def load_4season_depth_from_keyframe_data(
        keyframes_file_path: str
) -> DepthData:
    with open(keyframes_file_path, 'rt') as file:
        read_until_tag(file, KEYFRAME_TAG_CAM)
        fx, fy, cx, cy, width, height, *_ = (float(v) for v in file.readline().split(','))

        read_until_tag(file, KEYFRAME_TAG_DEPTH)
        # read depth and colors
        depth_coords = []
        while True:
            line1 = file.readline().strip()  # u,v,idepth_scaled,...
            _ = file.readline().strip()  # color
            if line1 == '':
                break
            line1 = [float(v) for v in line1.split(',')]
            depth_coords.append(line1[0:3])  # u, v, d

        depth_coords = np.array(depth_coords, dtype=kapture.RecordsDepth.dtype)
        depth_data = DepthData(f=(fx, fy), c=(cx, cy),
                               image_size=(int(height), int(width)),
                               coords=depth_coords)
        return depth_data


def convert_depth_data_to_depth_map(
        depth_data: DepthData,
        undefined_value=-1.0
) -> np.ndarray:
    """ converts depth map from 4 season format to kapture format """
    depth_map = np.ones(depth_data.image_size, dtype=depth_data.coords.dtype)
    depth_map *= undefined_value
    assert depth_data.coords.shape[1] == 3  # u, v, d
    for u, v, d in depth_data.coords:
        # convert d to metric according to given formula
        # d =  inverse depth value in 1/m
        depth_map[int(v), int(u)] = 1. / d if not math.isclose(d, 0.) else 0.
    return depth_map


def get_keyframe_file_names(
        keyframes_dir_path: str
) -> List[str]:
    """ populates keyframe directory and return dict with {filename: shot_id} """
    keyframe_filename_re = re.compile(r'^KeyFrame_(?P<shot_id>\d{19})\.txt$')
    filename_it = (path.basename(f) for f in glob(path.join(keyframes_dir_path, '*.txt')))
    filename_it = (filename for filename in filename_it
                   if keyframe_filename_re.match(filename))
    filename_it = (keyframe_filename_re.match(filename) for filename in filename_it)
    filename_it = {filename[0]: filename[1] for filename in filename_it}
    return filename_it


def load_gnss_poses_file(
    poses_file_path: str,
    sensor_id: str
) -> kapture.Trajectories:
    """ load GNSSPoses.txt to a kapture trajectories """
    pose_table = np.loadtxt(poses_file_path, delimiter=',', skiprows=1)
    timestamps_ns = pose_table[:, 0].astype(int)
    season_poses = pose_table[:, 1:8]
    trajectories = kapture.Trajectories()
    for timestamp_ns, (tx, ty, tz, qx, qy, qz, qw) in zip(timestamps_ns, season_poses):
        timestamp_ns = int(timestamp_ns)
        car_from_world = kapture.PoseTransform(r=[qw, qx, qy, qz], t=[tx, ty, tz]).inverse()
        trajectories[timestamp_ns, sensor_id] = car_from_world
    return trajectories


def load_transformations_file(
    transformations_file_path: str
) -> Tuple[Dict[str, kapture.PoseTransform], float]:
    """
    loads Transformations.txt into a dict [name] -> kapture.PoseTransform
    where
     - transform_S_AS: is from SLAM internal scale to metric scale.
     - TS_cam_imu is from IMU to the camera.
     - transform_w_gpsw: is from local GPS world (ENU) to visual world.
     - transform_gps_imu: is from IMU to GPS.
     - transform_e_gpsw: is from local GPS world (ENU) to global Earth frame (ECEF).
     - GNSS scale: scalar
     - Translation vector values correspond to x, y, z components of translation.
     - Rotation quaternion values correspond to q_x, q_y, q_z, w components of quaternion.

    frame notations:
     - gpsw = The Scene East-North-Up (Scene ENU) coordinate system

    """
    with open(transformations_file_path, 'rt') as transformations_file:
        lines = (line.strip() for line in transformations_file.readlines())
        lines = (line for line in lines if line)  # prune empty lines
        lines = list(lines)
    names = [line.split(':')[0][2:] for line in lines if line.startswith('#')]
    arrays = [np.fromstring(line, sep=',') for line in lines if not line.startswith('#')]
    transformations = {}
    for name, array in zip(names, arrays):
        if 'GNSS scale' == name:
            gnss_scale = array[0]
        else:
            tx, ty, tz, qx, qy, qz, qw = array.tolist()
            transformations[name] = kapture.PoseTransform(r=[qw, qx, qy, qz], t=[tx, ty, tz])
    return transformations, gnss_scale


def import_4seasons_trajectory(
        poses_file_path: str,
        transformations_file_path: str
) -> kapture.Trajectories:
    """
    imports trajectories.
    Using the globally optimized poses (inside GNSSPoses.txt),
    and transforming them using the chain of matrix multiplication from the Transformations.txt

    :param poses_file_path: full path to GNSSPoses.txt (name included)
    :param transformations_file_path: full path to Transformations.txt (name included)
    :return:
    """
    logger.info('importing images')
    # load gnss optimized poses
    trajectories = load_gnss_poses_file(poses_file_path, sensor_id=RIG_ID)
    # transforming them using the chain of matrix multiplication from the Transformations.txt
    # transform_e_gpsw @ np.linalg.inv(transform_w_gpsw) @ transform_S_AS @ scale_mat
    #   - AS: SLAM internal scale
    #   - S: metric scale
    #   - w: visual world
    #   - gpsw: local GPS world (ENU)
    #   - e: global Earth frame (ECEF).

    transformations, gnss_scale = load_transformations_file(transformations_file_path)
    transformations['transformations_gpsw_w'] = transformations['transform_w_gpsw'].inverse()
    transformations_ecef_from_slam = kapture.PoseTransform.compose([
        transformations['transform_e_gpsw'],
        transformations['transformations_gpsw_w'],
        transformations['transform_S_AS']
    ])
    # change for the 4season pose def: cam -> world (kapture is wolrd -> cam)
    poses = trajectories.inverse()
    kapture.trajectory_rescale_inplace(trajectories=trajectories, scale=gnss_scale)
    kapture.trajectory_transform_inplace(poses, pose_transform_pre=transformations_ecef_from_slam)
    # switch back to kapture (world -> cam)
    trajectories = poses.inverse()
    return trajectories


def import_4seasons_depth(
        keyframes_dir_path: str,
        kapture_dir_path: str,
        shot_id_to_timestamp: Dict[str, int],
        intrinsics: kapture.Camera
) -> (kapture.Sensors, kapture.RecordsDepth):
    """
    imports depth maps

    :param keyframes_dir_path:
    :param kapture_dir_path:
    :param shot_id_to_timestamp:
    :param intrinsics: input kapture camera (used to build depth sensor intrinsics)
    :return:
    """

    logger.info('importing depth maps')
    sensors = kapture.Sensors()
    # copy intrinsics
    sensors[DEPTH_ID] = kapture.Camera(
        sensor_type='depth',  name=DEPTH_ID, camera_type=intrinsics.camera_type, camera_params=intrinsics.camera_params)
    records_depth_maps = kapture.RecordsDepth()

    # make sure depth maps hosting directory exists
    depth_records_path = kapture.io.records.get_depth_map_fullpath(kapture_dir_path)
    os.makedirs(depth_records_path, exist_ok=True)
    #                                   KeyFrame_1602074967051661568.txt
    filename_it = get_keyframe_file_names(keyframes_dir_path)
    hide_progress_bar = getLogger().getEffectiveLevel() > logging.INFO
    for keyframe_filename, shot_id in tqdm(filename_it.items(), disable=hide_progress_bar):
        assert shot_id in shot_id_to_timestamp
        timestamp_ns = shot_id_to_timestamp[shot_id]
        keyframes_file_path = path.join(keyframes_dir_path, keyframe_filename)
        season_depth_data = load_4season_depth_from_keyframe_data(keyframes_file_path)
        # depth
        depth_map_file_name = path.join('undistorted_images', 'depth', f'{shot_id}.depth')
        records_depth_maps[timestamp_ns, DEPTH_ID] = depth_map_file_name
        depth_map = convert_depth_data_to_depth_map(season_depth_data)
        depth_map_file_path = kapture.io.records.get_depth_map_fullpath(kapture_dir_path, depth_map_file_name)
        kapture.io.records.depth_map_to_file(depth_map_file_path, depth_map)

    return sensors, records_depth_maps


def import_4seasons_imu(
        imu_file_path: str,
        shot_id_to_timestamp: Dict[str, int]
) -> (kapture.Sensors, kapture.RecordsAccelerometer, kapture.RecordsGyroscope):
    sensors = kapture.Sensors()
    sensors[ACCELEROMETER_ID] = kapture.Sensor(
        sensor_type='accelerometer', name=ACCELEROMETER_ID)
    sensors[GYROSCOPE_ID] = kapture.Sensor(
        sensor_type='gyroscope', name=GYROSCOPE_ID)

    accelerometer = kapture.RecordsAccelerometer()
    gyroscope = kapture.RecordsGyroscope()
    # Each line is specified as frame_id, (angular velocity (w_x, w_y, w_z), and linear acceleration (a_x, a_y, a_z)).
    # 1602074877342319360 -0.009163 0.018326 -0.070250 0.189211 0.860048 9.657110
    data = np.loadtxt(imu_file_path)
    shot_ids = data[:, 0].astype(np.int).astype(str)
    rotation_speeds = data[:, 1:4]
    translation_accels = data[:, 4:7]
    for shot_id, (rx, ry, rz), (ax, ay, az) in zip(shot_ids, rotation_speeds, translation_accels):
        timestamp = shot_id_to_timestamp.get(shot_id, guess_timestamp(shot_id))
        accelerometer[timestamp, ACCELEROMETER_ID] = kapture.RecordAccelerometer(ax, ay, az)
        gyroscope[timestamp, GYROSCOPE_ID] = kapture.RecordGyroscope(rx, ry, rz)
    return sensors, accelerometer, gyroscope


def import_4seasons_sequence(
        calibration_dir_path: str,
        recording_dir_path: str,
        kapture_dir_path: str,
        images_import_method: TransferAction,
        force_overwrite_existing: bool):
    """
    converts a 4 seasons recorded sequence (eg. recording_2020-10-07_14-47-51) to kapture format.

    :param calibration_dir_path: path to input calibration directory
    :param recording_dir_path: path to input sequence directory (e.g. recording_2020-10-07_14-47-51)
    :param kapture_dir_path: path to output kapture directory
    :param images_import_method:
    :param force_overwrite_existing:
    :return:
    """
    delete_existing_kapture_files(kapture_dir_path, force_erase=force_overwrite_existing)
    os.makedirs(kapture_dir_path, exist_ok=True)

    """
    recording_dir_path contains :
    KeyFrameData: contains the KeyFrameFiles.
    distorted_images: contains both the distorted images from the left and right camera, respectively.
    undistorted_images: contains both the undistorted images from the left and right camera, respectively.
    GNSSPoses.txt: is a list of 7DOF globally optimized poses (include scale from VIO to GNSS frame) for all keyframes
    (after GNSS fusion and loop closure detection). Each line is specified as frame_id, translation (t_x, t_y, t_z),
     rotation as quaternion (q_x, q_y, q_z, w), scale, fusion_quality (not relevant), and v3 (not relevant).
    Transformations.txt: defines transformations between different coordinate frames.
    imu.txt: contains raw IMU measurements. Each line is specified as frame_id, (angular velocity (w_x, w_y, w_z),
    and linear acceleration (a_x, a_y, a_z)).
    result.txt: contains the 6DOF visual interial odometry poses for every frame (not optimized).
    Each line is specified as timestamp (in seconds), translation (t_x, t_y, t_z), rotation as
    quaternion (q_x, q_y, q_z, w).
    septentrio.nmea: contains the raw GNSS measurements in the NMEA format.
    times.txt is a list of times in unix timestamps (in seconds), and exposure times (in milliseconds)
    for each frame (frame_id, timestamp, exposure).
    """

    # sensors
    sensors, rigs = load_4seasons_cameras(calibration_dir_path=calibration_dir_path)
    imported_kapture = kapture.Kapture(sensors=sensors, rigs=rigs)

    # timestamps
    times_filename = path.join(recording_dir_path, 'times.txt')
    shot_id_to_timestamp = load_times_ids(times_filename)

    # images
    records_camera = import_4seasons_images(
        recording_dir_path=recording_dir_path,
        kapture_dir_path=kapture_dir_path,
        shot_id_to_timestamp=shot_id_to_timestamp,
        sensors=sensors,
        images_import_method=images_import_method
    )
    imported_kapture.records_camera = records_camera

    # trajectories from keyframes
    poses_file_path = path.join(recording_dir_path, 'GNSSPoses.txt')
    transformations_file_path = path.join(recording_dir_path, 'Transformations.txt')
    trajectories = import_4seasons_trajectory(
        poses_file_path=poses_file_path,
        transformations_file_path=transformations_file_path
    )
    imported_kapture.trajectories = trajectories

    # depth maps from keyframes
    keyframes_dir_path = path.join(recording_dir_path, 'KeyFrameData')
    depth_sensors, depth_maps = import_4seasons_depth(
        keyframes_dir_path=keyframes_dir_path,
        kapture_dir_path=kapture_dir_path,
        shot_id_to_timestamp=shot_id_to_timestamp,
        intrinsics=imported_kapture.sensors[MASTER_CAM_ID]
    )
    imported_kapture.records_depth = depth_maps
    sensors.update(depth_sensors)
    # todo: add depth sensor to rig

    # imu.txt to accel and gyro
    imu_file_path = path.join(recording_dir_path, 'imu.txt')
    imu_sensors, records_accelerometer, records_gyroscope = import_4seasons_imu(
        imu_file_path=imu_file_path,
        shot_id_to_timestamp=shot_id_to_timestamp)
    sensors.update(imu_sensors)
    imported_kapture.records_accelerometer = records_accelerometer
    imported_kapture.records_gyroscope = records_gyroscope
    for imu_id in imu_sensors:
        imported_kapture.rigs[RIG_ID, imu_id] = kapture.PoseTransform()
        # TODO: set the orientation of imu into the rig.

    # GNSS data
    nmea_file_path = path.join(recording_dir_path, 'septentrio.nmea')
    gnss_sensors, records_gnss = extract_gnss_from_nmea(
        nmea_file_path=nmea_file_path, gnss_id='GNSS'
    )
    sensors.update(gnss_sensors)
    imported_kapture.records_gnss = records_gnss
    gnss_id = next(iter(gnss_sensors))
    imported_kapture.rigs[RIG_ID, gnss_id] = kapture.PoseTransform()

    # finally save the kapture csv files
    kapture_to_dir(kapture_dir_path, imported_kapture)


def import_4seasons_command_line() -> None:
    """
    Imports 4seasons dataset and save them as kapture using the parameters given on the command line.
    It assumes images are undistorted.
    """
    parser = argparse.ArgumentParser(
        description='Imports 4seasons dataset files to the kapture format.')
    parser_verbosity = parser.add_mutually_exclusive_group()
    parser_verbosity.add_argument(
        '-v', '--verbose', nargs='?', default=logging.WARNING, const=logging.INFO,
        action=kapture.utils.logging.VerbosityParser,
        help='verbosity level (debug, info, warning, critical, ... or int value) [warning]')
    parser_verbosity.add_argument(
        '-q', '--silent', '--quiet', action='store_const', dest='verbose', const=logging.CRITICAL)
    parser.add_argument('-f', '-y', '--force', action='store_true', default=False,
                        help='Force delete output if already exists.')
    # import ###########################################################################################################
    parser.add_argument('-i', '--input', required=True,
                        help='input path to 4 seasons record directory (e.g. ./recording_2020-10-07_14-47-51)')
    parser.add_argument('-c', '--calibration', required=False,
                        help='input path to 4 seasons calibration. If not given, assumed to be alongside input.')
    parser.add_argument('--image_transfer', type=TransferAction, default=TransferAction.link_absolute,
                        help=f'How to import images [link_absolute], '
                             f'choose among: {", ".join(a.name for a in TransferAction)}')
    parser.add_argument('-o', '--output', required=True, help='output directory.')
    ####################################################################################################################
    args = parser.parse_args()

    logger.setLevel(args.verbose)
    if args.verbose <= logging.DEBUG:
        # also let kapture express its logs
        kapture.utils.logging.getLogger().setLevel(args.verbose)

    recording_dir_path = path.abspath(args.input)
    calibration_dir_path = args.calibration or path.join(recording_dir_path, '..', 'calibration')
    calibration_dir_path = path.abspath(calibration_dir_path)

    import_4seasons_sequence(
        calibration_dir_path=calibration_dir_path,
        recording_dir_path=recording_dir_path,
        kapture_dir_path=args.output,
        images_import_method=args.image_transfer,
        force_overwrite_existing=args.force)


if __name__ == '__main__':
    import_4seasons_command_line()



#!/usr/bin/env python3
# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
script to easily  download kapture datasets.
"""

import argparse
import logging
import sys
import os
import os.path as path
import requests
import yaml
import fnmatch
from typing import Dict, Optional, List, Set
from tqdm import tqdm
import textwrap
import path_to_kapture  # noqa: F401
# import kapture
import kapture.utils.logging
from kapture.converter.downloader.download import download_file, get_remote_file_size
from kapture.converter.downloader.archives import untar_file, compute_sha256sum
from kapture.io.csv import get_version_from_csv_file
from kapture.utils.upgrade import upgrade_1_0_to_1_1_inplace, upgrade_1_0_to_1_1_orphan_features

logger = logging.getLogger('downloader')
logging.basicConfig(format='%(levelname)-8s::%(name)s: %(message)s')

INDEX_FILENAME = 'kapture_dataset_index.yaml'
INSTALL_LOG_FILENAME = 'kapture_dataset_installed.yaml'
LICENSE_LOG_FILENAME = 'kapture_dataset_licenses.yaml'
DEFAULT_DATASET_PATH = path.normpath(path.abspath('.'))
DEFAULT_REPOSITORY_URL = 'https://github.com/naver/kapture/raw/main/dataset'

# absolute path to root directory where all datasets are installed, its set for the all runtime
install_dirpath = None


def ask_confirmation(question):
    """
    ask "question" to the user.
    The "answer" return value is True for "yes" or False for "no".
    """
    validate = ['yes', 'y', 'ye']
    prompt = ' [y/N]\n'
    sys.stdout.write(question + prompt)
    user_choice = input().lower()
    return user_choice in validate


class InstallDir:
    """ Manage access to dataset installation dir and files into  """
    def __init__(
            self,
            index_filepath: str,
            install_dir_path: str):
        self._index_filepath = index_filepath
        self._install_dir_path = install_dir_path
        self._dataset_index_filepath = path.join(install_dir_path, INDEX_FILENAME)
        self._dataset_installed_list_filepath = path.join(install_dir_path, INSTALL_LOG_FILENAME)
        self._license_agreed_list_filepath = path.join(install_dir_path, LICENSE_LOG_FILENAME)

        self._dataset_index = None
        self._installed_index_cache = None  # list of already installed

    @property
    def install_root_path(self) -> str:
        return self._install_dir_path

    def _load_installed_index(self):
        """ load installed index from file. Index is empty if file does not exists. """
        logger.debug(f'loading {self._dataset_installed_list_filepath}')
        if path.isfile(self._dataset_installed_list_filepath):
            with open(self._dataset_installed_list_filepath, 'rt') as f:
                installed_index = set(yaml.safe_load(f))
        else:
            installed_index = set()
        self._installed_index_cache = installed_index

    def _write_installed_index(self):
        """ write installed index to file """
        logger.debug(f'writing {self._dataset_installed_list_filepath}')
        with open(self._dataset_installed_list_filepath, 'wt') as f:
            yaml.dump(list(self._installed_index), f)

    # Installed index
    @property
    def _installed_index(self) -> Set[str]:
        if self._installed_index_cache is None:
            # first time, need to be loaded
            self._load_installed_index()
        assert self._installed_index_cache is not None
        return self._installed_index_cache

    def mark_as_installed(self, dataset_name: str, installed: bool = True):
        _ = self._installed_index  # make sure its up to date
        if not self.is_installed(dataset_name) and installed:
            logger.debug(f'dataset marked installed {dataset_name} ')
            self._installed_index_cache.add(dataset_name)
            self._write_installed_index()
        if self.is_installed(dataset_name) and not installed:
            logger.debug(f'dataset marked uninstalled {dataset_name} ')
            self._installed_index_cache.discard(dataset_name)
            self._write_installed_index()

    def is_installed(self, dataset_name: str) -> bool:
        return dataset_name in self._installed_index

    # dataset index
    def load_datasets_from_file(
            self,
            filter_patterns: Optional[List[str]] = None
    ):
        """
        Parses and load data from the index files, under yaml format.
        the yaml file looks like :
        ----
        robotcar_seasons_02:
          url: http://download.europe.naverlabs.com//kapture/robotcar_seasons_02.tar.gz
          sha256sum: 542ef47c00d5e387cfb0dcadb2459ae2fb17d59010cc51bae0c49403b4fa6a18
          license: https://filesamples.com/samples/document/txt/sample3.txt # optional
        ----

        :param filter_patterns: optional input list of unix-like patterns (e.g. SiLDa*) to filter datasets
        :return: dict name -> [url, sub_path, sha256sum]
        """
        if not path.isfile(self._index_filepath):
            raise FileNotFoundError('no index file: do an update.')
        with open(self._index_filepath, 'rt') as f:
            datasets_yaml = yaml.safe_load(f)

        if len(datasets_yaml) == 0:
            raise FileNotFoundError('invalid index file: do an update.')

        nb_total = len(datasets_yaml)
        # filter only dataset matching filter_patterns
        if filter_patterns:
            datasets_yaml = {dataset_name: data_yaml
                             for dataset_name, data_yaml in datasets_yaml.items()
                             if any(fnmatch.fnmatch(dataset_name, pattern) for pattern in filter_patterns)}

        logger.debug(f'will prob status for {len(datasets_yaml)}/{nb_total} datasets ...')
        datasets = {}
        hide_progress_bar = True or logger.getEffectiveLevel() > logging.INFO
        for dataset_name, data_yaml in tqdm(datasets_yaml.items(), disable=hide_progress_bar):
            datasets[dataset_name] = Dataset(name=dataset_name,
                                             install_dir=self,
                                             archive_url=data_yaml['url'],
                                             archive_sha256sum=data_yaml['sha256sum'],
                                             install_script_filename=data_yaml.get('install_script'),
                                             license_url=data_yaml.get('license'))

        return datasets

    def read_licenses(self) -> Set[str]:
        """ returns the list of licences already agreed by user """
        if not path.isfile(self._license_agreed_list_filepath):
            # not licences yet
            return set()

        with open(self._license_agreed_list_filepath, 'rt') as file:
            license_list = set(yaml.safe_load(file))
        return license_list

    def update_licenses(self, licenses_list_updt: Set[str]):
        """ updates the list licences by adding the givens ones  """
        licenses_list = self.read_licenses()
        licenses_list.update(licenses_list_updt)
        with open(self._license_agreed_list_filepath, 'wt') as file:
            yaml.dump(list(licenses_list), file)


class Dataset:
    def __init__(
            self,
            name: str,
            install_dir: InstallDir,
            archive_url: str,
            archive_sha256sum: str,
            install_script_filename: Optional[str] = None,
            license_url: Optional[str] = None
    ):
        """
        :param name: name of the archive (dataset or part of a dataset)
        :param install_dir: root directory where all datasets are installed.
        :param archive_url: remote url of the dataset archive (tar).
        :param archive_sha256sum: expected sha256 sum of the archive file.
        :param install_script_filename: if given, this script is to be called to finish installation (eg. dl 3rd party).
        :param license_url: if given, url to license file to agree on.
        """
        self._name = name

        self._install_dir = install_dir
        self._archive_filepath = path.join(install_dir.install_root_path, name + '.tar.gz')
        self._archive_url = archive_url
        self._archive_sha256sum_remote = archive_sha256sum
        self._install_script_filename = install_script_filename
        # license
        self._license_url = license_url

    def mark_as_installed(self, installed=True):
        self._install_dir.mark_as_installed(self._name, installed)

    def is_installed(self) -> bool:
        return self._install_dir.is_installed(self._name)

    def is_sha256_consistent(self) -> bool:
        """ check sha256sum of the archive against the expected sha256. Returns true if they are the same. """
        if not path.isfile(self._archive_filepath):
            return False
        # size is consistent, check sha256
        logger.debug(f'checking sha256sum of {path.basename(self._archive_filepath)}, may take a while...')
        sha256sum_archive_local = compute_sha256sum(self._archive_filepath)
        if sha256sum_archive_local != self._archive_sha256sum_remote:
            logger.warning(f'sha256sum discrepancy for {self._archive_filepath} :\n'
                           f'\tlocal :{sha256sum_archive_local}\n'
                           f'\tremote:{self._archive_sha256sum_remote}')
            return False
        logger.debug(f'{path.basename(self._archive_filepath)} is valid (sha256sum)')
        return True

    @property
    def url(self) -> str:
        return self._archive_url

    @property
    def license_url(self) -> Optional[str]:
        return self._license_url

    def prob_status(self, check_online=False) -> str:
        """
        gives the actual dataset status
         - not installed: means is not installed (wo info about the server)
         - online: means not installed, not downloaded, but reachable.
         - not reachable: means not installed, not downloaded, but NOT reachable.
         - installed: means has been downloaded and installed (extracted).
         - downloaded: means has been downloaded (tar) but not installed (extracted) yet.
         - incomplete: means partially downloaded
         - corrupted: means that the downloaded archive is corrupted (inconsistent size or sh256).
         - unknown: should not happen

        :param check_online: If true, will ping the url to check its actually online
        """

        probing_status = None
        size_archive_local = size_archive_online = -1
        if self.is_installed():
            # yaml file says its installed, trust it (no other choices).
            probing_status = 'installed'

        if probing_status is None and not path.isfile(self._archive_filepath):
            # not installed, no archive there, check its online
            if not check_online:
                probing_status = 'not installed'
            else:
                remote_file_size_in_bytes = get_remote_file_size(url=self._archive_url)
                if remote_file_size_in_bytes is not None:
                    probing_status = f'online {remote_file_size_in_bytes / (1024 * 1024):5.0f} MB '
                else:
                    probing_status = 'not reachable'

        # not installed, but archive there, check 1) its incomplete or 2) corrupted. If neither, is just downloaded.
        if probing_status is None:
            assert path.isfile(self._archive_filepath)
            # 1) check its incomplete: has it proper size ?
            size_archive_online = get_remote_file_size(self._archive_url)
            size_archive_local = int(path.getsize(self._archive_filepath))
            if size_archive_online is None:
                logger.critical('impossible to retrieve remote file size.')
                probing_status = 'corrupted'
            elif size_archive_local > size_archive_online:
                logger.critical(f'inconsistent file size, file is bigger than it is supposed to be'
                                f' ({size_archive_online} vs {size_archive_local}).')
                probing_status = 'corrupted'
            elif size_archive_local < size_archive_online:
                logger.debug(f'file_size_online={size_archive_online} != file_size_local={size_archive_local}')
                probing_status = 'incomplete'

        if probing_status is None:
            # 2) size is consistent, check sha256
            assert size_archive_local == size_archive_online
            if not self.is_sha256_consistent():
                probing_status = 'corrupted'

        # archive is there, not installed, with valid sha256 then it must be just downloaded.
        if probing_status is None:
            probing_status = 'downloaded'

        return probing_status

    def __repr__(self) -> str:
        return f'{self.prob_status():10} | {self._name:30} | {self._archive_url}'

    def download(self,
                 force_overwrite: bool = False,
                 nb_attempt: int = 2,
                 previous_status=None) -> str:
        status = previous_status or self.prob_status()
        if force_overwrite and path.isfile(self._archive_filepath):
            logger.debug('remove previously downloaded file and start from scratch.')
            os.remove(self._archive_filepath)
            status = self.prob_status()

        if status == 'downloaded':
            logger.info(f'{path.basename(self._archive_filepath)} is already downloaded.')
            return status

        for attempt in range(nb_attempt):
            if status == 'downloaded':
                logger.debug(f'{path.basename(self._archive_filepath)} is downloaded')
                break
            # check archive file integrity
            if status == 'corrupted':
                logger.warning(f'archive {path.basename(self._archive_filepath)} is corrupted. '
                               f'It will be downloaded again.')
                # if corrupted: remove the archive and start over
                os.remove(self._archive_filepath)
            logger.debug(f'downloading {path.basename(self._archive_filepath)} (attempt {attempt + 1}/{nb_attempt})')
            download_file(self._archive_url, self._archive_filepath)
            status = self.prob_status()
        return status

    def install(self,
                force_overwrite: bool = False,
                no_cleaning: bool = False) -> str:
        """
        Install handles download and untar, and possibly tar remove and launch of an install script.

        """
        # make sure self._status is up to date.
        if force_overwrite:
            self.mark_as_installed(False)

        status = self.prob_status()
        if status == 'installed':
            logger.info(f'{self._install_dir.install_root_path} already exists: skipped')
            return status

        # download (lets try it twice
        status = self.download(previous_status=status)

        if status != 'downloaded':
            # if still not downloaded, it means it failed to download
            logger.critical(f'failed to download {self._name}')
            return status

        # deflating
        logger.debug(f'deflating {path.basename(self._archive_filepath)} to {self._install_dir.install_root_path}')
        untar_file(self._archive_filepath, self._install_dir.install_root_path)

        # optionally: script ?
        if self._install_script_filename is not None:
            logger.debug(f'installation script  {self._install_script_filename}')
            install_script_filepath = path.join(self._install_dir.install_root_path, self._install_script_filename)
            logger.debug(f'applying installation script {install_script_filepath}')
            os.system(install_script_filepath)
        # optionally: clean tar.gz file
        if not no_cleaning:
            logger.debug(f'removing archive file {path.basename(self._archive_filepath)}')
            os.remove(self._archive_filepath)

        # done
        self.mark_as_installed()
        self.upgrade()
        logger.debug(f'done installing {self._name}')
        status = self.prob_status()
        return status

    def upgrade(self):
        # upgrade ?
        status = self.prob_status()
        if status == 'installed':
            # list all sensors.txt files
            file_list = [os.path.join(dp, f) for dp, dn, filenames in os.walk(self._install_dir.install_root_path)
                         for f in filenames if f == 'sensors.txt']
            upgrade_successful = False
            # check version
            for filename in file_list:
                version = get_version_from_csv_file(filename)
                if version is None or version == '1.0':
                    logger.debug(f'upgrading {filename} from version {version} to 1.1')
                    kapture_path = path.dirname(path.dirname(filename))
                    # run inplace upgrade script
                    upgrade_1_0_to_1_1_inplace(kapture_path, None, None, None, 'L2', 'L2')
                    logger.info(f'{kapture_path} - successfully upgraded to 1.1')
                elif version == '1.1':
                    logger.info(f'{filename} already in 1.1, no need to upgrade')
                else:
                    logger.warning(f'{filename} - version {version} is unknown')
            # attempt upgrade of orphan features
            csv_feature_names = ['keypoints.txt', 'descriptors.txt', 'global_features.txt']
            file_list = [os.path.join(dp, f) for dp, dn, filenames in os.walk(self._install_dir.install_root_path)
                         for f in filenames if f in csv_feature_names]
            orphan_lfeat_paths = set()
            orphan_gfeat_paths = set()
            for filename in file_list:
                version = get_version_from_csv_file(filename)
                if version is None or version == '1.0':
                    filename_c = path.abspath(filename).replace('\\', '/').rstrip('/')
                    filename_split = filename_c.split('/')
                    if len(filename) > 1 and \
                            (filename_split[-1] == 'keypoints.txt' or filename_split[-1] == 'descriptors.txt') and \
                            (filename_split[-2] == 'keypoints' or filename_split[-2] == 'descriptors'):
                        orphan_lfeat_paths.add('/'.join(filename_split[0:-2]))
                    elif len(filename) > 1 and \
                            filename_split[-1] == 'global_features.txt' and \
                            filename_split[-2] == 'global_features':
                        orphan_gfeat_paths.add('/'.join(filename_split[0:-2]))
            try:
                if len(orphan_lfeat_paths) > 0 or len(orphan_gfeat_paths) > 0:
                    logger.info(f'upgrade orphan features for {orphan_lfeat_paths} {orphan_gfeat_paths}')
                    upgrade_1_0_to_1_1_orphan_features(list(orphan_lfeat_paths), list(orphan_gfeat_paths))
                upgrade_successful = True
            except Exception as e:
                logger.warning(f'upgrade_1_0_to_1_1_orphan_features {e} for {orphan_lfeat_paths} {orphan_gfeat_paths}')
            return upgrade_successful
        else:
            logger.warning('dataset not yet installed, cannot attempt upgrade')
            return False


def check_licenses(
        install_dir: InstallDir,
        datasets: Dict[str, Dataset]
):
    """
    Check user is aware of licences.
    expected behavior:
    - license == url,
    - only ask user once per url,
    - dont ask user for licenses already approved (and registered in a dedicated file),
    - If user disagree a license, prevent corresponding datasets to be installed.

    :param install_dir:
    :param datasets:
    :return:
    """

    # regroup licenses
    license_datasets = {}  # license_url -> [list of datasets referring to that license]
    for name, dataset in datasets.items():
        if dataset.license_url is not None:
            license_datasets.setdefault(dataset.license_url, list())
            license_datasets[dataset.license_url].append(name)
    logger.debug(f'dataset refer {len(license_datasets)} licences.')

    # retrieve already agreed licenses.
    licenses_already_agreed = install_dir.read_licenses()
    # and filter only the ones we are concerned about.
    licenses_already_agreed = {
        license_url
        for license_url in licenses_already_agreed
        if license_url in license_datasets
    }

    # removes from license_datasets, the the ones already agreed upon.
    for license_to_skip in licenses_already_agreed:
        del license_datasets[license_to_skip]

    if not license_datasets:
        # no license to agree, job done
        return

    # some licenses remains to be approved
    logger.info(f'{len(license_datasets)} to be approved:')
    licenses_already_agreed = set()
    # prompt user for each one
    for license_url, dataset_names in license_datasets.items():
        r = requests.get(license_url, allow_redirects=True)
        if r.status_code != requests.codes.ok:
            raise ConnectionError(f'unable to grab {license_url} (code:{r.status_code})')
        logger.info(f'here after the license terms for : {", ".join(dataset_names)}')
        wrapper = textwrap.TextWrapper(width=80)
        print(wrapper.fill(text=r.text))
        print('----------------------')
        print(f'from: {license_url}')
        agree = ask_confirmation('do you agree on terms ?')
        if not agree:
            # in case of disagreement: remove all dataset with that license,
            licenses_already_agreed.add(license_url)
            for name_of_disagreement in dataset_names:
                logger.critical(f'removing dataset {name_of_disagreement} due to license disagreement.')
                del datasets[name_of_disagreement]

    logger.debug(f'update list of agreed licences.')
    for license_to_skip in licenses_already_agreed:
        del license_datasets[license_to_skip]
    install_dir.update_licenses(license_datasets)


def kapture_download_dataset(args, index_filepath: str):
    """
    Do the dataset download command
    """
    global_status = 0
    install_dir = InstallDir(index_filepath=index_filepath, install_dir_path=args.install_path)
    if args.cmd == 'update':
        logger.info(f'updating dataset list from {args.repo} ...')
        index_remote_url = args.repo + '/' + INDEX_FILENAME
        logger.debug(f'retrieving index at {index_remote_url}')
        r = requests.get(index_remote_url, allow_redirects=True)
        if r.status_code != requests.codes.ok:
            raise ConnectionError(f'unable to grab {index_remote_url} (code:{r.status_code})')
        with open(index_filepath, 'wt') as f:
            f.write(r.text)
        datasets = install_dir.load_datasets_from_file()
        logger.info(f'dataset index retrieved successfully: {len(datasets)} datasets')

    elif args.cmd == 'list':
        logger.info(f'listing dataset {index_filepath} ...')
        datasets = install_dir.load_datasets_from_file(filter_patterns=args.dataset)
        for name, dataset in datasets.items():
            status = dataset.prob_status(check_online=args.full)
            if status == "not reachable" or status == "incomplete" or status == "corrupted":
                global_status = 1
            print(f'{status:^16}| {name:40} | {dataset.url}')

    elif args.cmd == 'install':
        logger.debug(f'will install dataset: {args.dataset} ...')
        datasets = install_dir.load_datasets_from_file(filter_patterns=args.dataset)
        if len(datasets) == 0:
            raise ValueError('There is no matching dataset.'
                             ' Make sure you used quotes (") to prevent shell interpreting * wildcard.')

        # ask for license agreement
        check_licenses(install_dir=install_dir, datasets=datasets)

        logger.info(f'{len(datasets)} dataset will be installed.')
        for name, dataset in datasets.items():
            logger.info(f'{name}: starting installation  ...')
            status = dataset.install(force_overwrite=args.force, no_cleaning=args.no_cleaning)
            logger.info(f'{name} install: ' + 'successful' if status == 'installed' else 'failed')

    elif args.cmd == 'upgrade':
        logger.debug(f'will install dataset: {args.dataset} ...')
        datasets = install_dir.load_datasets_from_file(filter_patterns=args.dataset)
        if len(datasets) == 0:
            raise ValueError('There is no matching dataset.'
                             ' Make sure you used quotes (") to prevent shell interpreting * wildcard.')

        logger.info(f'{len(datasets)} dataset will be installed.')
        for name, dataset in datasets.items():
            logger.info(f'{name}: starting installation  ...')
            dataset.upgrade()

    elif args.cmd == 'download':
        logger.debug(f'will download dataset: {args.dataset} ...')
        datasets = install_dir.load_datasets_from_file(filter_patterns=args.dataset)
        if len(datasets) == 0:
            raise ValueError('There is no matching dataset.'
                             ' Make sure you used quotes (") to prevent shell interpreting * wildcard.')
        logger.info(f'{len(datasets)} dataset will be downloaded.')
        for name, dataset in datasets.items():
            logger.info(f'downloading {name} ...')
            dataset.download(force_overwrite=args.force)

    return global_status


def kapture_download_dataset_cli():
    """
    Parse the kapture_download_dataset command line .
    """
    parser = argparse.ArgumentParser(description='download kapture datasets.')
    parser_verbosity = parser.add_mutually_exclusive_group()
    parser_verbosity.add_argument(
        '-v', '--verbose', nargs='?', default=logging.INFO, const=logging.INFO,
        action=kapture.utils.logging.VerbosityParser,
        help='verbosity level (debug, info, warning, critical, ... or int value) [warning]')
    parser_verbosity.add_argument(
        '-q', '--silent', '--quiet', action='store_const', dest='verbose', const=logging.CRITICAL)
    parser.add_argument('--install_path', default=path.normpath(path.join(DEFAULT_DATASET_PATH)),
                        help=f'path to index files listing all datasets'
                             f' [{path.normpath(path.join(DEFAULT_DATASET_PATH))}]')
    subparsers = parser.add_subparsers(help='sub-command help', dest='cmd')
    ####################################################################################################################
    # create the parser for the "update" command
    parser_update = subparsers.add_parser('update', help='update dataset index')
    parser_update.set_defaults(cmd='update')
    parser_update.add_argument('--repo', default=DEFAULT_REPOSITORY_URL,
                               help='url of the repository.')
    ####################################################################################################################
    parser_list = subparsers.add_parser('list', help='display dataset index')
    parser_list.set_defaults(cmd='list')
    parser_list.add_argument('dataset', nargs='*', default=[])
    parser_list.add_argument('--full', action='store_true', default=False,
                             help='Display status and size of remote dataset.')
    ####################################################################################################################
    parser_install = subparsers.add_parser('install', help='install dataset')
    parser_install.set_defaults(cmd='install')
    parser_install.add_argument('-f', '--force', action='store_true', default=False,
                                help='Force installation even if dataset has already been installed.')
    parser_install.add_argument('--no_cleaning', action='store_true', default=False,
                                help='Do not delete downloaded tar.gz file.')
    parser_install.add_argument('dataset', nargs='*', default=[],
                                help='name of the dataset to download. Can use unix-like wildcard.')
    ####################################################################################################################
    parser_download = subparsers.add_parser('download', help='download dataset, without installing it')
    parser_download.set_defaults(cmd='download')
    parser_download.add_argument('-f', '--force', action='store_true', default=False,
                                 help='Force installation even if dataset has already been installed.')
    parser_download.add_argument('dataset', nargs='*', default=[],
                                 help='name of the dataset to download. Can use unix-like wildcard.')
    ####################################################################################################################
    parser_install = subparsers.add_parser('upgrade', help='upgrade dataset to 1.1')
    parser_install.set_defaults(cmd='upgrade')
    parser_install.add_argument('dataset', nargs='*', default=[],
                                help='name of the dataset to download. Can use unix-like wildcard.')
    ####################################################################################################################

    args = parser.parse_args()

    logger.setLevel(args.verbose or logging.INFO)
    logger.debug(f'{sys.argv[0]} \\\n' + '  \\\n'.join(
        '--{:20} {:100}'.format(k, str(v))
        for k, v in vars(args).items()))

    try:
        index_filepath = path.join(args.install_path, INDEX_FILENAME)
        if not args.cmd:
            parser.print_help()
            logger.critical(f'Choose command among [ {" | ".join(subparsers.choices)} ]')
            exit(-1)
        return kapture_download_dataset(args, index_filepath)

    except Exception as e:
        logger.critical(e)
        if logger.getEffectiveLevel() >= logging.DEBUG:
            raise e
        return -1


if __name__ == '__main__':
    sys.exit(kapture_download_dataset_cli())



#!/usr/bin/env python3
# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Convert kapture data in version 1.0 to version 1.1 (only orphan features)
"""

import logging
import argparse

import path_to_kapture  # noqa: F401
import kapture.utils.logging
from kapture.utils.upgrade import upgrade_1_0_to_1_1_orphan_features

logger = logging.getLogger('upgrade_1_0_to_1_1_orphan_features')


def upgrade_1_0_to_1_1_orphan_features_command_line() -> None:
    """
    upgrade orphan features to kapture 1.1. Orphan features are features stored outside the kapture folder
    they must follow the kapture-localization recommendation
    https://github.com/naver/kapture-localization/blob/main/doc/tutorial.adoc#recommended-dataset-structure
    """
    parser = argparse.ArgumentParser(
        description='convert kapture data in version 1.0 to version 1.1 (only orphan features)')
    parser_verbosity = parser.add_mutually_exclusive_group()
    parser_verbosity.add_argument(
        '-v', '--verbose', nargs='?', default=logging.WARNING, const=logging.INFO,
        action=kapture.utils.logging.VerbosityParser,
        help='verbosity level (debug, info, warning, critical, ... or int value) [warning]')
    parser_verbosity.add_argument(
        '-q', '--silent', '--quiet', action='store_const', dest='verbose', const=logging.CRITICAL)
    # export ###########################################################################################################
    parser.add_argument('-gfeat', '--global-features', nargs='+', default=[],
                        help=('paths to the global features folders '
                              'examples dataset/global_features/apgem dataset/global_features/delg'))
    parser.add_argument('-lfeat', '--local-features', nargs='+', default=[],
                        help=('paths to the local features folders '
                              'examples dataset/local_features/r2d2 dataset/local_features/d2_tf'))
    args = parser.parse_args()

    logger.setLevel(args.verbose)
    if args.verbose <= logging.DEBUG:
        # also let kapture express its logs
        kapture.utils.logging.getLogger().setLevel(args.verbose)

    upgrade_1_0_to_1_1_orphan_features(args.local_features, args.global_features)


if __name__ == '__main__':
    upgrade_1_0_to_1_1_orphan_features_command_line()
